from pathlib import Path
from xml.etree import ElementTree

from pydantic import validate_call

import kmm


class Header(kmm.FunctionalBase):
    car_direction: kmm.CarDirection
    position: int
    sync: int

    @staticmethod
    @validate_call
    def from_path(path: Path, raise_on_malformed_data: bool = True):
        """
        Loads header data from .hdr file.
        """
        try:
            tree = ElementTree.parse(path)
        except ElementTree.ParseError:
            if raise_on_malformed_data:
                raise ValueError("Unable to parse header file, invalid XML.")
            tree = attempt_to_patch_xml(path)

        position, sync = kmm.header.position_sync(tree)
        return Header(
            position=position,
            sync=sync,
            car_direction=kmm.header.car_direction(tree),
        )


def attempt_to_patch_xml(path: Path):
    """
    Attempts to patch a broken header file by adding missing end tags.
    """
    lines = path.read_text().splitlines()
    start_tags = []
    for line in lines:
        if line.startswith("</") and line.endswith(">"):
            try:
                index = max(
                    i
                    for i, start_tag in enumerate(start_tags)
                    if start_tag == line[2:-1]
                )
            except ValueError:
                raise ElementTree.ParseError(f"Missing start tag for {line}")
            start_tags.pop(index)
        elif line.startswith("<") and line.endswith(">") and not line.startswith("<?"):
            start_tags.append(line[1:-1])
    for start_tag in start_tags[::-1]:
        lines.append(f"</{start_tag}>")
    try:
        return ElementTree.ElementTree(ElementTree.fromstring("\n".join(lines)))
    except ElementTree.ParseError as e:
        raise ValueError("Unable to parse header file, invalid XML.") from e


def test_header():
    Header.from_path("tests/ascending_B.hdr")


def test_empty_header():
    Header.from_path("tests/empty.hdr")



import re
from xml.etree import ElementTree


def position_sync(tree: ElementTree):

    root = tree.getroot()
    sync_tags = [child.text for child in root if child.tag == "Sync"]

    if len(sync_tags) == 0:
        raise ValueError("Did not find a Sync tag in header.")

    sync_tag = sync_tags[0]

    position = re.search(
        r"Position = \"(\d*)\"",
        sync_tag,
    )
    if position is not None:
        position = int(position.group(1))
    else:
        raise ValueError("""Did not find a "Position" field under the Sync tag.""")

    sync = re.search(
        r"Sync = \"(\d*)\"",
        sync_tag,
    )

    if sync is not None:
        sync = int(sync.group(1))
    else:
        raise ValueError("""Did not find a "Sync" field under the Sync tag.""")

    return position, sync



from pydantic import BaseModel, ConfigDict


class FunctionalBase(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    def map(self, fn, *args, **kwargs):
        return fn(self, *args, **kwargs)

    def replace(self, **kwargs):
        new_dict = self.model_dump()
        new_dict.update(**kwargs)
        return type(self)(**new_dict)



from pathlib import Path

import pandas as pd
from pydantic import ConfigDict, validate_call

import kmm
from kmm.header.header import Header


class Positions(kmm.FunctionalBase):
    dataframe: pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @staticmethod
    @validate_call
    def from_path(
        path: Path,
        raise_on_malformed_data: bool = True,
        replace_commas: bool = True,
    ):
        """
        Loads positions from .kmm or .kmm2 file.
        """
        if path.suffix == ".kmm":
            dataframe = kmm.positions.read_kmm(path)
        elif path.suffix == ".kmm2":
            dataframe = kmm.positions.read_kmm2(
                path,
                raise_on_malformed_data=raise_on_malformed_data,
                replace_commas=replace_commas,
            )
        else:
            raise ValueError(f"Unable to parse file type {path.suffix}")

        return Positions(dataframe=dataframe)

    @staticmethod
    @validate_call
    def read_sync_adjust(
        kmm_path: Path,
        header_path: Path,
        adjustment: kmm.PositionAdjustment = kmm.PositionAdjustment.WIRE_CAMERA,
        raise_on_malformed_data: bool = True,
        replace_commas: bool = True,
    ):
        """
        Loads positions from .kmm or .kmm2 file + .hdr file, then performs
        frame index sync, position adjustment and geodetic coordinate transformation.
        """
        header = kmm.Header.from_path(header_path, raise_on_malformed_data)
        return (
            Positions.from_path(
                kmm_path,
                raise_on_malformed_data=raise_on_malformed_data,
                replace_commas=replace_commas,
            )
            .sync_frame_index(header, adjustment, raise_on_malformed_data)
            .geodetic()
        )

    @validate_call
    def sync_frame_index(
        self,
        header: Header,
        adjustment: kmm.PositionAdjustment,
        raise_on_malformed_data: bool = True,
    ):
        return kmm.positions.sync_frame_index(
            self, header, adjustment, raise_on_malformed_data
        )

    def geodetic(self):
        return kmm.positions.geodetic(self)


def test_read_kmm():
    positions = Positions.read_sync_adjust(
        "tests/ascending_B.kmm", "tests/ascending_B.hdr"
    )
    assert len(positions.dataframe) > 0


def test_read_kmm2():
    positions = Positions.read_sync_adjust(
        "tests/ascending_B.kmm2", "tests/ascending_B.hdr"
    )
    assert len(positions.dataframe) > 0


def test_empty_kmm():
    positions = Positions.from_path("tests/empty.kmm")
    assert len(positions.dataframe) == 0


def test_empty_kmm2():
    positions = Positions.from_path("tests/empty.kmm2")
    assert len(positions.dataframe) == 0



from io import StringIO
from pathlib import Path

import numpy as np
import pandas as pd
from pydantic import validate_call


@validate_call
def read_kmm(path: Path, replace_commas: bool = True):
    try:
        if replace_commas:
            with open(path, "r", encoding="latin1") as f:
                content = f.read()
            content = content.replace(",", ".")
            file_obj = StringIO(content)
        else:
            file_obj = path
        return pd.read_csv(
            file_obj,
            sep="\t",
            encoding="latin1" if not replace_commas else None,
            names=[
                "centimeter",
                "track_section",
                "kilometer",
                "meter",
                "track_lane",
                "1?",
                "2?",
                "3?",
                "4?",
                "5?",
                "northing",
                "easting",
                "8?",
                "9?",
            ],
            dtype=dict(
                track_section=str,
                kilometer=np.int32,
                meter=np.int32,
                track_lane=str,
                northing=np.float32,
                easting=np.float32,
            ),
        )
    except Exception as e:
        raise ValueError("Unable to parse kmm2 file, invalid csv.") from e



import numpy as np
from pydantic import validate_call

from kmm import CarDirection, PositionAdjustment
from kmm.header.header import Header
from kmm.positions.positions import Positions


@validate_call(config=dict(arbitrary_types_allowed=True))
def sync_frame_index(
    positions: Positions,
    header: Header,
    adjustment: PositionAdjustment,
    raise_on_malformed_data: bool = True,
):
    if raise_on_malformed_data:
        validate_meter_increments(positions)
    frame_index = (
        (positions.dataframe["centimeter"].values + header.position - header.sync) / 10
    ).astype(int)

    if header.car_direction == CarDirection.A:
        dataframe = positions.dataframe.iloc[:-adjustment].assign(
            frame_index=frame_index[adjustment:]
        )
    elif header.car_direction == CarDirection.B:
        dataframe = positions.dataframe.iloc[adjustment:].assign(
            frame_index=frame_index[:-adjustment]
        )
    else:
        raise ValueError(f"Unsupported car direction {header.car_direction}")

    return positions.replace(
        dataframe=dataframe.assign(car_direction=header.car_direction)
    )


def validate_meter_increments(positions):
    for (track_section, kilometer), group in positions.dataframe.groupby(
        ["track_section", "kilometer"]
    ):
        diffs = np.sign(group["meter"].values[1:] - group["meter"].values[:-1])
        if len(diffs) >= 10 and (diffs > 0).mean() < 0.9 and (diffs < 0).mean() < 0.9:
            raise ValueError(
                f"Inconsistent directions at track_section {track_section}, kilometer {kilometer}."
            )


def test_sync_frame_index_ascending_B_kmm():
    from kmm import Header

    synced = sync_frame_index(
        Positions.from_path("tests/ascending_B.kmm"),
        Header.from_path("tests/ascending_B.hdr"),
        PositionAdjustment.WIRE_CAMERA,
    ).dataframe
    assert synced[synced["frame_index"] == 675]["kilometer"].values[0] == 292
    assert synced[synced["frame_index"] == 675]["meter"].values[0] == 737 + 8


def test_sync_frame_index_ascending_B_kmm2():
    from kmm import Header

    synced = sync_frame_index(
        Positions.from_path("tests/ascending_B.kmm2"),
        Header.from_path("tests/ascending_B.hdr"),
        PositionAdjustment.WIRE_CAMERA,
    ).dataframe
    assert synced[synced["frame_index"] == 5]["kilometer"].values[0] == 292
    assert synced[synced["frame_index"] == 5]["meter"].values[0] == 802 + 8


def test_sync_frame_index_ascending_A():
    from kmm import Header

    synced = sync_frame_index(
        Positions.from_path("tests/ascending_A.kmm2"),
        Header.from_path("tests/ascending_A.hdr"),
        PositionAdjustment.WIRE_CAMERA,
    ).dataframe
    assert synced[synced["frame_index"] == 8]["kilometer"].values[0] == 534
    assert synced[synced["frame_index"] == 8]["meter"].values[0] == 336 - 8


def test_sync_frame_index_descending_A():
    from kmm import Header

    synced = sync_frame_index(
        Positions.from_path("tests/descending_A.kmm2"),
        Header.from_path("tests/descending_A.hdr"),
        PositionAdjustment.WIRE_CAMERA,
    ).dataframe
    assert synced[synced["frame_index"] == 1000]["kilometer"].values[0] == 57
    assert synced[synced["frame_index"] == 1000]["meter"].values[0] == 2 + 8
